''' PodWorker | modules | logging.py '''

import os
from dotenv import load_dotenv


env_path = os.getcwd() + '/.env'
load_dotenv(env_path)  # Load environment variables


def log(message, level='INFO'):
    '''
    Log message to stdout
    '''
    if os.environ.get('DEBUG', 'False') == 'true':
        print(f'{level} | {message}')


def log_secret(secret_name, secret, level='INFO'):
    '''
    Censors secrets for logging
    Replaces everything except the first and last characters with *
    '''
    if secret is None:
        secret = 'Could not read environment variable.'
        log(f"{secret_name}: {secret}", 'ERROR')
    else:
        secret = str(secret)
        redacted_secret = secret[0] + '*' * len(secret) + secret[-1]
        log(f"{secret_name}: {redacted_secret}", level)


log('Logging module loaded')

log_secret('AI_API_KEY', os.environ.get('AI_API_KEY', None))
log_secret('WEBHOOK_GET_WORK', os.environ.get('WEBHOOK_GET_WORK', None))
log_secret('WEBHOOK_POST_OUTPUT', os.environ.get('WEBHOOK_POST_OUTPUT', None))

log_secret('BUCKET_ENDPOINT_URL', os.environ.get('BUCKET_ENDPOINT_URL', None))
log_secret('BUCKET_ACCESS_KEY_ID', os.environ.get('BUCKET_ACCESS_KEY_ID', None))
log_secret('BUCKET_SECRET_ACCESS_KEY', os.environ.get('BUCKET_SECRET_ACCESS_KEY', None))



'''
PodWorker | modules | inference.py
Interacts with the model to make predictions.
'''

# -------------------------- Import Model Predictors ------------------------- #
from infer import Predictor

from .logging import log


class Model:
    ''' Interface for the model.'''

    def __init__(self):
        '''
        Loads the model.
        '''
        self.predictor = Predictor()
        self.predictor.setup()
        log('Model loaded.')

    def input_validation(self, model_inputs):
        '''
        Validates the input.
        '''
        input_types = self.predictor.inputs(model_inputs)

        for key, value in model_inputs.items():
            if key not in input_types:
                raise ValueError(f'Input {key} not expected.')

            if not isinstance(value, input_types[key]):
                raise ValueError(f'Input {key} should be {input_types[key]}.')

    def run(self, model_inputs):
        '''
        Predicts the output of the model.
        '''
        return self.predictor.run(model_inputs)



''' PodWorker | jobs.py '''

import os
import time
import json
import requests

from .logging import log
from . import upload, inference


def get(worker_id):
    '''
    Get next job from job endpoint, returns job json.
    The job format is:
    {
        "id": {job_id},
        "input": {job_input}
    }
    '''
    if os.environ.get('WEBHOOK_GET_WORK', None) is None:
        log('WEBHOOK_GET_WORK not set, switching to get_local', 'WARNING')
        return get_local()

    get_work_url = str(os.environ.get('WEBHOOK_GET_WORK')
                       ).replace('$ID', worker_id)

    log(f"Requesting job from {get_work_url}")

    headers = {"Authorization": f"{os.environ.get('AI_API_KEY')}"}

    try:
        assigned_job = requests.get(
            get_work_url,
            headers=headers,
            timeout=180
        )
    except requests.exceptions.Timeout:
        log("Timeout while requesting job", 'WARNING')
        return None

    if assigned_job.status_code == 200:
        log(f"TAKE_JOB URL response: {assigned_job.status_code}")
        return assigned_job.json()

    if assigned_job.status_code == 204:
        log(f"TAKE_JOB URL response: {assigned_job.status_code}")
        return None

    log(f"TAKE_JOB URL response: {assigned_job.status_code}", 'ERROR')
    return None


def run(job_id, job_input):
    '''
    Run the job.
    Returns list of URLs and Job Time
    '''
    time_job_started = time.time()

    log(f"Started working on {job_id} at {time_job_started} UTC")

    model = inference.Model()
    job_output = model.run(job_input)

    object_url = upload.upload_image(job_id, job_output["image"])
    job_output["image"] = object_url

    job_duration = time.time() - time_job_started
    job_duration_ms = int(job_duration * 1000)

    return job_output, job_duration_ms


def post(worker_id, job_id, job_output, job_time):
    '''
    Complete the job.
    '''
    if os.environ.get('WEBHOOK_POST_OUTPUT', None) is None:
        return

    job_data = {
        "output": job_output
    }

    job_data = json.dumps(job_data, ensure_ascii=False)

    job_done_url = str(os.environ.get('WEBHOOK_POST_OUTPUT'))
    job_done_url = job_done_url.replace('$ID', job_id)
    job_done_url = job_done_url.replace('$RUNPOD_POD_ID', worker_id)

    headers = {
        "charset": "utf-8",
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": f"{os.environ.get('AI_API_KEY')}"
    }

    try:
        requests.post(job_done_url, data=job_data, headers=headers, timeout=10)
    except requests.exceptions.Timeout:
        log(f"Timeout while completing job {job_id}")

    log(f"Completed job {job_id} in {job_time} ms")

    return


def error(worker_id, job_id, error_message):
    '''
    Report an error to the job endpoint, marking the job as failed.
    '''
    log(f"Reporting error for job {job_id}: {error_message}", 'ERROR')

    job_output = {
        "error": error_message
    }

    job_output = json.dumps(job_output, ensure_ascii=False)

    job_error_url = str(os.environ.get('WEBHOOK_POST_OUTPUT'))
    job_error_url = job_error_url.replace('$ID', job_id)
    job_error_url = job_error_url.replace('$RUNPOD_POD_ID', worker_id)

    headers = {
        "charset": "utf-8",
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": f"{os.environ.get('AI_API_KEY')}"
    }

    try:
        requests.post(job_error_url, data=job_output, headers=headers, timeout=10)
    except requests.exceptions.Timeout:
        log(f"Timeout while erroring job {job_id}")


# ------------------------------- Local Testing ------------------------------ #
def get_local():
    '''
    Returns contents of test_inputs.json
    '''
    if not os.path.exists('test_inputs.json'):
        return None

    with open('test_inputs.json', 'r', encoding="UTF-8") as file:
        return json.loads(file.read())



''' Tests for runpod | serverless| modules | download.py '''
# pylint: disable=R0903,W0613

import unittest
from unittest.mock import patch, mock_open

from runpod.serverless.modules.download import download_input_objects


def mock_requests_get(*args, **kwargs):
    '''
    Mocks requests.get
    '''
    class MockResponse:
        ''' Mocks requests.get response '''

        def __init__(self, content, status_code):
            '''
            Mocks requests.get response
            '''
            self.content = content
            self.status_code = status_code

    if args[0] == 'https://example.com/picture.jpg':
        return MockResponse(b'nothing', 200)

    return MockResponse(None, 404)


class TestDownloadInputObjects(unittest.TestCase):
    ''' Tests for download_input_objects '''

    @patch('os.makedirs', return_value=None)
    @patch('requests.get', side_effect=mock_requests_get)
    @patch('builtins.open', new_callable=mock_open)
    def test_download_input_objects(self, mock_open_file, mock_get, mock_makedirs):
        '''
        Tests download_input_objects
        '''
        objects = download_input_objects(
            ['https://example.com/picture.jpg', ]
        )

        self.assertEqual(len(objects), 1)
        self.assertIn('https://example.com/picture.jpg', mock_get.call_args_list[0][0])
        mock_open_file.assert_called_once_with(objects[0], 'wb')
        mock_makedirs.assert_called_once_with('input_objects', exist_ok=True)



'''
PodWorker | modules | lifecycle.py

Performs the following lifecycle operations:
- Shutting down the worker
'''
# pylint: disable=too-few-public-methods

import os
import uuid
import threading
import requests

from .logging import log


class LifecycleManager:
    ''' Lifecycle Manager '''

    def __init__(self):
        '''
        Initialize Lifecycle Manager
        '''
        self.worker_id = os.environ.get('RUNPOD_POD_ID', str(uuid.uuid4()))
        log(f'Worker ID: {self.worker_id}')

        self.job_id = None

    def heartbeat_ping(self):
        '''
        Pings the heartbeat endpoint
        '''
        webhook_ping = os.environ.get('WEBHOOK_PING', None)
        ping_interval = int(os.environ.get('PING_INTERVAL', 10000))

        if self.job_id is not None:
            ping_params = {
                'job_id': self.job_id,
            }
        else:
            ping_params = {}

        if webhook_ping is not None:
            webhook_ping = webhook_ping.replace('$RUNPOD_POD_ID', self.worker_id)
            requests.get(webhook_ping, params=ping_params, timeout=ping_interval/1000)

        log(f'Heartbeat sent to {webhook_ping} interval: {ping_interval}ms params: {ping_params}')

        threading.Timer(ping_interval/1000, self.heartbeat_ping).start()



'''
PodWorker | modules | download.py

Called when inputs are images or zip files.
Downloads them into a temporary directory called "input_objects".
This directory is cleaned up after the job is complete.
'''

import os
import uuid
import requests


def download_input_objects(object_locations):
    '''
    Cycles through the object locations and downloads them.
    Returns the list of downloaded objects paths.
    '''
    os.makedirs('input_objects', exist_ok=True)

    objects = []
    for object_url in object_locations:
        response = requests.get(object_url, timeout=5)

        file_name = os.path.basename(object_url)
        file_extension = os.path.splitext(file_name)[1]

        object_name = f'{uuid.uuid4()}{file_extension}'

        with open(f'input_objects/{object_name}', 'wb') as output_file:
            output_file.write(response.content)

        objects.append(f'input_objects/{object_name}')

    return objects



''' PodWorker | modules | upload.py '''

import os
import time
from io import BytesIO

from PIL import Image
from boto3 import session
from botocore.config import Config

from .logging import log

# --------------------------- S3 Bucket Connection --------------------------- #
bucket_session = session.Session()

boto_config = Config(
    signature_version='s3v4',
    retries={
        'max_attempts': 3,
        'mode': 'standard'
    }
)

if os.environ.get('BUCKET_ENDPOINT_URL', None) is not None:
    boto_client = bucket_session.client(
        's3',
        endpoint_url=os.environ.get('BUCKET_ENDPOINT_URL', None),
        aws_access_key_id=os.environ.get('BUCKET_ACCESS_KEY_ID', None),
        aws_secret_access_key=os.environ.get('BUCKET_SECRET_ACCESS_KEY', None),
        config=boto_config
    )
else:
    boto_client = None  # pylint: disable=invalid-name


# ---------------------------------------------------------------------------- #
#                                 Upload Image                                 #
# ---------------------------------------------------------------------------- #
def upload_image(job_id, job_results):
    '''
    Upload image to bucket storage.
    '''
    if boto_client is None:
        # Save the output to a file
        for index, result in enumerate(job_results):
            output = BytesIO()
            img = Image.open(result)
            img.save(output, format=img.format)

            os.makedirs("uploaded", exist_ok=True)
            with open(f"uploaded/{index}.png", "wb") as file_output:
                file_output.write(output.getvalue())

        return []

    object_urls = []
    for index, result in enumerate(job_results):
        output = BytesIO()
        img = Image.open(result)
        img.save(output, format=img.format)

        bucket = time.strftime('%m-%y')

        # Upload to S3
        boto_client.put_object(
            Bucket=f'{bucket}',
            Key=f'{job_id}/{index}.png',
            Body=output.getvalue(),
            ContentType="image/png"
        )

        output.close()

        presigned_url = boto_client.generate_presigned_url(
            'get_object',
            Params={
                'Bucket': f'{bucket}',
                'Key': f'{job_id}/{index}.png'
            }, ExpiresIn=604800)

        object_urls.append(presigned_url)

        log(f"Presigned URL generated: {presigned_url}")

    return object_urls



'''
runpod | serverless | pod_worker.py
Called to convert a container into a worker pod for the runpod serverless platform.
'''

import os
import shutil

from .modules import lifecycle, job
from .modules.logging import log


def start_worker():
    '''
    Starts the worker.
    '''
    worker_life = lifecycle.LifecycleManager()
    worker_life.heartbeat_ping()
    log("Worker lifecycle manager started.")

    while True:
        next_job = job.get(worker_life.worker_id)

        if next_job is not None:
            worker_life.job_id = next_job['id']

            try:
                if 'input' not in next_job:
                    log("No input provided. Erroring out request.", "ERROR")
                    job.error(worker_life.worker_id, next_job['id'], "No input provided.")
                else:
                    job_output, job_duration_ms = job.run(next_job['id'], next_job['input'])
                    job.post(worker_life.worker_id, next_job['id'], job_output, job_duration_ms)
            except (KeyError, ValueError, RuntimeError) as err:
                job.error(worker_life.worker_id, next_job['id'], str(err))
            finally:
                # -------------------------------- Job Cleanup ------------------------------- #
                shutil.rmtree("input_objects", ignore_errors=True)
                shutil.rmtree("output_objects", ignore_errors=True)

                if os.path.exists('output.zip'):
                    os.remove('output.zip')

                worker_life.job_id = None

        if os.environ.get('WEBHOOK_GET_WORK', None) is None:
            log("Local testing complete, exiting.")
            break
